/** @file
	Parser PgSQL driver: Configure file for autoconf-disabled platforms.

	Copyright (c) 2001-2012 Art. Lebedev Studio (http://www.artlebedev.com)

	Author: Alexander Petrosyan <paf@design.ru> (http://design.ru/paf)

	$Id: config_fixed.h,v 1.6 2012-03-16 10:05:04 moko Exp $


	tested only with MS Visual C++ V6
*/

#define HAVE_STDIO_H
#define HAVE_STDLIB_H
#define HAVE_STRING_H
#define HAVE_SETJMP_H
#define HAVE_CTYPE_H
#define HAVE_SETJMP_H
#define HAVE_PUTENV
